#include <Blackfin_definitions.h>
#include <functions.h>
#include <SDDefinitions.h>

int CommandBuffer [9]; 					// Buffers to store commands
int ResponseBuffer [17]; 				// Buffers to store responses
int InterruptCount;	 					// Count to keep track of the interrupts
volatile int Done80Cycles;				// Flag to indicate that 80 cycles needed for initialization are complete (SPI DMA interrupt)
volatile int DataBlockReceived; 		// Flag to indicate that data block is received using DMA
volatile int DataBlockTransmitted; 		// Flag to indicate that data block is transmitted using DMA
int Response;							// Response value from the SD Card.
int TokenFlag;							// Flag used to indicate that token value (0xFE) is received.
volatile int DataResponseFlag;			// Flag used to indicate that data block is written to SD successfully
int SPITransferStatus;					// Slow SPI transfer status
int SPITransferCount;					// Command / Data transfer count - Slow SPI mode
volatile int Command = 0;				// Command that is being sent to SD
char COREorDMA = 0;						// Core or DMA
char Dummy[12];							// Dummy Buffer
int N;									// Final SPI Transfer Count
char InterruptCause;					// Interrupt Cause that decides what part of the ISR is to be executed
int BusyCheck;							// Indicates that the SD card is released from the busy state after write
int DummyBuffer [17];					// Dummy buffer
extern int ReadTokenCount;				// Tracks the number of clock cycles sent to receive the token from card
int TempBuff;							// Temporary buffer to store the read value
extern volatile int ReadToken;				// Indicates that the token is received after Read

/*
Function Definitions
*/
//--------------------------------------------------------------------------//
// Function:	PLLSetupBF													//
//																			//
// Parameters:	None														//
//																			//
// Return:		None														//
//																			//
// Description:	This function sets up the VCO, Core Clock and System clock  //
//				for BF. These clock values are defined in BF_Parameters.h	//
//--------------------------------------------------------------------------//
void PLLSetupBF()
{
    *pSIC_IWR = 0x1;
	*pPLL_CTL = SET_MSEL(VCO / CLKin);
	ssync();
	idle();
	*pPLL_DIV = SET_SSEL(VCO / SystemCLK) | ((VCO_CoreCLK)&0x30);
	asm("ssync;");
}

//--------------------------------------------------------------------------//
// Function:	PinMuxing													//
//																			//
// Parameters:	None														//
//																			//
// Return:		None														//
//																			//
// Description:	This function sets up the multiplexed pins on BF52x for the //
//				required functionality.										//
//--------------------------------------------------------------------------//
void PinMuxing(void)
{
	*pPORTG_MUX = 0x2;
	*pPORTG_FER = 0x1C;
}
//--------------------------------------------------------------------------//
// Function:	SetUpInterruptsBF											//
//																			//
// Parameters:	None														//
//																			//
// Return:		None														//
//																			//
// Description:	This function sets up the interrupts for SPI 				//
//				The SPI interrupts are mapped to IVG10 						//
//--------------------------------------------------------------------------//
void SetUpInterruptsBF()
{
    register_handler(ik_ivg10,SDInitIsr);
    *pSIC_IMASK |= IRQ_DMA7; 
}

//--------------------------------------------------------------------------//
// Function:	SetUpSlaveSelects											//
//																			//
// Parameters:	None														//
//																			//
// Return:		None														//
//																			//
// Description:	This function configures PG1 as output and drives it High   //
//--------------------------------------------------------------------------//
void SetUpSlaveSelects()
{
    *pPORTGIO_DIR = PG1; 								//Configure PG1 as output - CS/
    *pPORTGIO_CLEAR = PG1;								//Drive PF1 Logic HIGH - SSS/ de-asserted state
}

//--------------------------------------------------------------------------//
// Function:	SDInit														//
//																			//
// Parameters:	None														//
//																			//
// Return:		None														//
//																			//
// Description:	This function initializes the SD card. It includes the 		//
//				initial 80 cycles, CMD0 and CMD1							//
//--------------------------------------------------------------------------//
void SDInit ()
{
    int i;
    InterruptCause = 2;
    Done80Cycles = 0;
    SetupDMA(12, Dummy);
    SetUpAndEnableSpiDma();
    while(!Done80Cycles);
        
    SetUpSlaveSelects();
    InterruptCause = 1;
    
    //CMD0
	Response = 1;
	while(Response)
	{
		SDCommand(CMD_GO_IDLE_STATE, 0);
		SDGetResponse(1);
	}
	Delay();	//8 clock cycles reqd before sending CMD1
   
	//CMD1	 	
	Response = 1;
	while(Response)
	{
	    SDCommand(CMD_SEND_OP_COND, 0);
		SDGetResponse(0);
		Delay();	
	}
}

//--------------------------------------------------------------------------//
// Function:	SDReadData													//
//																			//
// Parameters:	Start Address, destination									//
//																			//
// Return:		None														//
//																			//
// Description:	This function reads the specified location of the SD card  //
//				and stores the read value at Destination					//
//--------------------------------------------------------------------------//
void SDReadData (int Address, char* Destination)
{
    TokenFlag = 0;
    while(!TokenFlag)
    {
        int i;
    	InterruptCause = 6;
	    Response = 1;
	    
		//CMD17
		while(Response)
	    {
			SDCommand(CMD_READ_BLOCK,Address);		
			SDGetResponse(0);
			if(Response != 0)
			{
			    printf("Read Block command not successful\n");
				printf("Response from SD: 0x%x\n",Response);
			}
	    }
	    printf("Read Block command successful\n");
	   
	    ReadTokenCount = 0;		//Wait for the SD card to send token (0xFE)
		InterruptCause = 7;
		TokenFlag = 0;
	    SetUpAndEnableSpiCoreWrite();
	    ReadToken = 0;
	    while(!ReadToken);
    
    }
    
	InterruptCause = 2;			//Setup DMA and receive data
	DataBlockReceived = 0;
	SetupDMA(515, Destination);
    SetUpAndEnableSpiDma();
    while(!DataBlockReceived);    
}

//--------------------------------------------------------------------------//
// Function:	SDWriteData												//
//																			//
// Parameters:	Start Address, Source										//
//																			//
// Return:		None														//
//																			//
// Description:	This function writes the values from the sorce buffer into  //
//				specified location of the SD card 							//
//--------------------------------------------------------------------------//
void SDWriteData (int Address, char* Source)
{
    int i, temp;
    InterruptCause = 1;
    Response = 1;
    
    //CMD24
    while(Response)
	{
		SDCommand(CMD_WRITE_BLOCK,Address);
		SDGetResponse(0);
		Delay();	
		if(Response != 0)
		{
		    printf("Write command not successful\n");
			printf("Response from SD: 0x%x\n",Response);
		}
	}
	
	printf("Write command successful\n");
	
	InterruptCause = 4;
	DataBlockTransmitted = 0;

	SetupDMATransmit(515, Source); // Setup and start DMA for the transmit
    SetUpAndEnableSpiDmaTransmit();
    while(!DataBlockTransmitted);
    
    InterruptCause = 3;
    SetUpAndEnableSpiCore();
    BusyCheck = 0;
    
    ResponseBuffer[0] = *pSPI_RDBR;	
    while(!BusyCheck);
        
}

void SDEraseGroup (int StartSector, int NumberOfSectors)
{
    
    InterruptCause = 1;
    Response = 1;
    while(Response)
    {
		SDCommand(ERASE_GROUP_START, StartSector);
		SDGetResponse(0);
		Delay();
    }
    
	Response = 1;
	while(Response)
    {
		SDCommand(ERASE_GROUP_END, StartSector+NumberOfSectors);
		SDGetResponse(0);
		Delay();
    }
	
	Response = 1;
	while(Response)
	{
		SDCommand(CMD_ERASE, 0xF);
		SDGetResponse(0);
		Delay();
	}
	
	BusyCheck = 0;
	InterruptCause = 3;
    SetUpAndEnableSpiCore();
    
    ResponseBuffer[0] = *pSPI_RDBR;	
    while(!BusyCheck);
}

void StatusReport()
{
    InterruptCause = 1;
	Response = 1;
	SDCommand(SEND_STATUS, 0);   
}

void SendCSDValue()
{
    InterruptCause = 1;
	Response = 1;
	SDCommand(SEND_CSD, 0);  
}

void SDSetWriteProtect(int Address)
{
    InterruptCause = 1;
	Response = 1;
	SDCommand(SET_WRITE_PROT, 0);  
}

void SDClearWriteProtect(int Address)
{
    InterruptCause = 1;
	Response = 1;
	SDCommand(CLR_WRITE_PROT, 0);	  
}

void SetUpCommandArray(int Command, int value)
{
    // Filling up the command buffer
	CommandBuffer[0] = 0x40 | Command;
	CommandBuffer[1] = (value & 0xff000000) >> 24;
	CommandBuffer[2] = (value & 0xff0000) >> 16;
	CommandBuffer[3] = (value & 0xff00) >> 8;
	CommandBuffer[4] = value & 0xff;
	
	// CMD0 should be 0x40 00 00 00 00 95
	if(Command == 0)
		CommandBuffer[5] = 0x95;
	else
		CommandBuffer[5] = 0x01;
		
	CommandBuffer[6] = 0xff;
	CommandBuffer[7] = 0xff;
	CommandBuffer[8] = 0xff;
}

void Delay()
{
    SetUpAndEnableSpiCore();  
	SPITransferCount = 9;
	SPITransferStatus = 1;
    N = 16;
    ResponseBuffer[0] = *pSPI_RDBR;
    while (SPITransferStatus);
}

void SDCommand(int Command, int value)
{
    SPITransferCount = 0;
    SPITransferStatus = 1;
    N = 16;
    SetUpCommandArray(Command, value);
    SetUpAndEnableSpiCore();
    *pSPI_TDBR = CommandBuffer[0];  
    ResponseBuffer[0] = *pSPI_RDBR;
    while (SPITransferStatus);
    asm("nop;");
}


int  SDGetResponse (int Value)
{
    int i;
    for(i=0;i<17;i++)
	{
	    if(ResponseBuffer[i] == Value)
	    {
	        Response = 0;
	    }
	}
    return Response;
    
}

void SDSetBlockLength (int Length)
{
    Response = 1;
    while(Response!=0)
    {
		SDCommand(CMD_SET_BLOCKLEN,512);
		SDGetResponse(0);
		Delay();	
		if(Response != 0)
		{
	    	printf("Set Block Length command not successful\n");
			printf("Response from SD: 0x%x\n",Response);
			//exit(1);
		}
		
		printf("Set Block Length command successful\n");

	}
    
}
void SetUpAndEnableSpiCore ()
{
    *pSPI_CTL = 0;
    *pSPI_CTL = CPOL | MSTR |GM | CPHA;				//Configure BF SPI as Master, CPOL=1, CPHA = 1
    *pSPI_BAUD = SystemCLK / SPIBaudProcess/2;
    *pPORTGIO_CLEAR = PG1;									//Drive PF1 Logic LOW - Assert SSS/
    *pSPI_CTL |= SPE;
}

void SetUpAndEnableSpiCoreWrite ()
{
    *pSPI_CTL = 0;
    *pSPI_CTL = CPOL | MSTR |GM | CPHA | 0x1;				//Configure BF SPI as Master, CPOL=1, CPHA = 1
    *pSPI_BAUD = SystemCLK / SPIBaudProcess/2;
    *pPORTGIO_CLEAR = PG1;									//Drive PF1 Logic LOW - Assert SSS/
    *pSPI_CTL |= SPE;
}

void SetUpAndEnableSpiDma ()
{
    
    *pSPI_CTL = 0;
    *pSPI_CTL = CPOL | MSTR | SZ | GM | CPHA | RDBR_DMA;	//Configure BF SPI as Master, CPOL=1, CPHA = 1, DMA on RX read
    *pSPI_BAUD = SystemCLK / SPIBaudInit/2;
    //*pFIO_FLAG_S= PF1;										
    *pSPI_CTL |= SPE;
}


void SetUpAndEnableSpiDmaTransmit (void)
{
    *pSPI_CTL = 0;
    *pSPI_CTL = CPOL | MSTR | SZ | GM | CPHA | TDBR_DMA;	//Configure BF SPI as Master, CPOL=1, CPHA = 1, DMA on RX read
    *pSPI_BAUD = SystemCLK / SPIBaudInit/2;
    //*pFIO_FLAG_S= PF1;
     										
    *pSPI_CTL |= SPE;
}

void SetupDMA(int Count, char* StartAddr)
{
    *pDMA7_START_ADDR = StartAddr;
    *pDMA7_X_COUNT = Count;
    *pDMA7_X_MODIFY = 1;
    *pDMA7_CONFIG = FLOW_STOP | WNR | DI_EN | DMAEN;      
}

void SetupDMATransmit(int Count, char* StartAddr)
{
	*pDMA7_START_ADDR = StartAddr;
    *pDMA7_X_COUNT = Count;
    *pDMA7_X_MODIFY = 1;
    *pDMA7_CONFIG = FLOW_STOP | DI_EN | DMAEN;  	    
}

void DisableCoreSPI(void)
{
    while(!(*pSPI_STAT & 0x0020));				//Check for Rx Buffer Full
 	while(!(*pSPI_STAT & 0x0001));				//Check for SPIF
 	*pSPI_CTL &= ~SPE;							//Disable SPI
 	TempBuff = *pSPI_RDBR;			//Clears the SPI RX buffer full status
 	asm("nop;");
 	//*pFIO_FLAG_S = PF1;							//Drive PF1 Logic HIGH - de-assert SSS/
}	

void DisableDmaSPI()
{
    //*pFIO_FLAG_S = PF1;							
    *pDMA7_CONFIG = 0x0;
    *pSPI_CTL &= ~SPE;
}

void DisableCoreSPIWrite(void)
{
	*pSPI_CTL &= ~SPE;							//Disable SPI
	TempBuff = *pSPI_RDBR;			//Clears the SPI RX buffer full status
 	asm("nop;");
}

void DisableDmaSPITransmit(void)
{
    int temp;
    while(*pDMA7_IRQ_STATUS & DMA_RUN);
    while(!(*pSPI_STAT & 0x0001));
    for(temp = 0; temp<100000; temp++)
    {
        asm("nop;");
    }
    *pDMA7_CONFIG = 0x0;
    *pSPI_CTL &= ~SPE;
}
